import time
from datetime import timedelta

from django.utils import timezone

from celery_progress.backend import Progress

from aleksis.apps.evalu.models import EvaluationRegistration
from aleksis.core.celery import app
from aleksis.core.util.pdf import generate_pdf_from_template


@app.task
def generate_privacy_form_task(pk: int) -> bool:
    registration = EvaluationRegistration.objects.get(pk=pk)

    file_object, result = generate_pdf_from_template(
        "evalu/registration/print.html", {"object": registration}
    )
    progress = Progress(result)
    start_time = timezone.now()
    while timezone.now() - start_time < timedelta(minutes=5):
        info = progress.get_info()
        if info["complete"]:
            file_object.refresh_from_db()
            if file_object.file:
                registration.privacy_form.save("privacy_form.pdf", file_object.file.file)
                registration.save()
                return True
        time.sleep(0.5)
    return False



# Generated by Django 4.2.6 on 2023-10-31 12:03

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("evalu", "0016_customevaluationitem"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="evaluationitem",
            name="part_site_order_unique",
        ),
    ]



from base64 import b64decode, b64encode
from collections import OrderedDict
from typing import Dict, List, Optional, Union

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Count, Prefetch, Q, QuerySet
from django.forms import model_to_dict
from django.utils import timezone
from django.utils.functional import classproperty
from django.utils.translation import gettext as _

from ckeditor.fields import RichTextField
from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey

from aleksis.apps.cursus.models import Subject
from aleksis.apps.evalu.managers import (
    EvaluationGroupManager,
    EvaluationGroupQuerySet,
    EvaluationPhaseManager,
    EvaluationPhaseQuerySet,
    EvaluationRegistrationManager,
    EvaluationRegistrationQuerySet,
)
from aleksis.core.mixins import ExtensibleModel, ExtensiblePolymorphicModel
from aleksis.core.models import Group, GroupType, Person
from aleksis.core.util.core_helpers import get_site_preferences


class EvaluationPhase(ExtensibleModel):
    name = models.CharField(max_length=255, verbose_name=_("Display Name"))
    evaluated_group = models.ForeignKey(
        to=Group,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        verbose_name=_("Group with teachers which can register for evaluation"),
    )

    evaluation_group_types = models.ManyToManyField(
        to=GroupType,
        null=True,
        blank=True,
        verbose_name=_("Groups with these types can evaluate persons"),
    )

    registration_date_start = models.DateField(
        verbose_name=_("First date teachers can register themselves for evaluation")
    )
    registration_date_end = models.DateField(
        verbose_name=_("Last date teachers can register themselves for evaluation")
    )
    evaluation_date_start = models.DateField(
        verbose_name=_("First date teachers can start the evaluation")
    )
    evaluation_date_end = models.DateField(verbose_name=_("Date when all evaluations stop"))

    privacy_notice = RichTextField(verbose_name=_("Privacy notice which teachers have to agree"))

    objects = EvaluationPhaseManager.from_queryset(EvaluationPhaseQuerySet)()

    class Meta:
        verbose_name = _("Evaluation phase")
        verbose_name_plural = _("Evaluation phases")

    def __str__(self):
        return self.name

    def clean(self):
        if self.registration_date_end < self.registration_date_start:
            raise ValidationError(
                _(
                    "The start of the registration period must "
                    "be before the end of the registration period."
                )
            )

        if self.evaluation_date_end < self.evaluation_date_start:
            raise ValidationError(
                _(
                    "The start of the evaluation period must "
                    "be before the end of the evaluation period."
                )
            )

    @property
    def registration_running(self):
        now_dt = timezone.now().date()
        return self.registration_date_start <= now_dt <= self.registration_date_end

    @property
    def status(self) -> str:
        now_dt = timezone.now().date()
        if self.evaluation_date_start <= now_dt <= self.evaluation_date_end:
            return "evaluation"
        elif self.evaluation_date_end < now_dt:
            return "evaluation_closed"
        else:
            return "not_started"

    @property
    def members_with_registration(self) -> QuerySet:
        return self.evaluated_group.members.all().prefetch_related(
            Prefetch(
                "evaluation_registrations",
                queryset=EvaluationRegistration.objects.filter(phase=self),
            )
        )

    def delete_data(self):
        for registration in self.registrations.filter(delete_after_phase=True, deleted=False):
            registration.delete_data()


class ComparisonGroup(ExtensibleModel):
    phase = models.ForeignKey(
        to=EvaluationPhase,
        on_delete=models.CASCADE,
        verbose_name=_("Evaluation phase"),
        related_name="comparison_groups",
    )
    name = models.CharField(max_length=255, verbose_name=_("Name"))
    groups = models.ManyToManyField(
        Group, verbose_name=_("Groups"), related_name="evalu_comparison_groups"
    )

    class Meta:
        verbose_name = _("Comparison group")
        verbose_name_plural = _("Comparison groups")

    def __str__(self):
        return self.name

    def is_valid_to_store(self, subject: Subject) -> bool:
        """Check if it's allowed to store data for this comparison group."""
        minimum_number = get_site_preferences()["evalu__number_of_persons_comparison_groups"]
        return (
            Person.objects.filter(
                owner_of__in=Group.objects.filter(subject_id=subject.id).filter(
                    Q(pk__in=self.groups.all()) | Q(parent_groups__in=self.groups.all())
                )
            )
            .distinct()
            .count()
            >= minimum_number
        )

    def is_valid_to_show(self, subject: Subject) -> bool:
        """Check if this group is allowed to be shown to the user."""
        minimum_number_persons = get_site_preferences()[
            "evalu__number_of_persons_comparison_groups"
        ]
        minimum_number_results = get_site_preferences()[
            "evalu__number_of_results_comparison_groups"
        ]
        number_of_results = self.results.filter(subject=subject).count()
        number_of_persons = (
            self.done_evaluations_comparison.filter(evaluation_group__group__subject_id=subject.id)
            .values_list("evaluation_group__registration", flat=True)
            .count()
        )
        return (
            number_of_results >= minimum_number_results
            and number_of_persons >= minimum_number_persons
        )


class EvaluationKeyPair(models.Model):
    private_key = models.TextField(verbose_name=_("Private key"), editable=False)
    public_key = models.TextField(verbose_name=_("Public key"), editable=False)

    class Meta:
        verbose_name = _("Evaluation key set")
        verbose_name_plural = _("Evaluation key set")

    def __str__(self):
        return f"Key {self.pk}"

    def get_public_key(self) -> Optional[RSAPublicKey]:
        """Get the public key."""
        if not self.public_key:
            return None
        public_key = serialization.load_pem_public_key(
            self.public_key.encode(), backend=default_backend()
        )
        return public_key

    def get_private_key(self, password: str) -> Optional[RSAPrivateKey]:
        """Get the private key."""
        if not self.private_key:
            return None
        private_key = serialization.load_pem_private_key(
            self.private_key.encode(), password=password.encode(), backend=default_backend()
        )
        return private_key

    @classmethod
    def create(cls, password: str) -> "EvaluationKeyPair":
        """Create a new public/private key pair from a given password."""
        pair = cls()
        # Generate a key
        private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )
        public_key = private_key.public_key()

        # Store the keys
        pair.private_key = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.BestAvailableEncryption(password.encode()),
        ).decode()
        pair.public_key = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        ).decode()
        pair.save()

        return pair

    def encrypt(self, message: Union[str, bytes]) -> str:
        """Encrypt a message with the public key."""
        public_key = self.get_public_key()
        if not isinstance(message, bytes):
            message = message.encode()
        ciphertext = public_key.encrypt(
            message,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()), algorithm=hashes.SHA256(), label=None
            ),
        )
        return b64encode(ciphertext).decode()

    def decrypt(self, ciphertext: Union[str, bytes], password: [str, RSAPrivateKey]) -> bytes:
        """Decrypt a message with the private key."""
        if isinstance(password, RSAPrivateKey):
            private_key = password
        else:
            private_key = self.get_private_key(password)

        if not isinstance(ciphertext, bytes):
            ciphertext = ciphertext.encode()
        plaintext = private_key.decrypt(
            b64decode(ciphertext),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()), algorithm=hashes.SHA256(), label=None
            ),
        )
        return plaintext

    def test(self, password: str) -> bool:
        """Test to unlock the private key."""
        self.get_private_key(password)
        return True


class EvaluationRegistration(ExtensibleModel):
    phase = models.ForeignKey(
        to=EvaluationPhase,
        on_delete=models.CASCADE,
        verbose_name=_("Evaluation phase"),
        related_name="registrations",
    )
    person = models.ForeignKey(
        to=Person,
        on_delete=models.CASCADE,
        verbose_name=_("Evaluated person"),
        related_name="evaluation_registrations",
    )
    privacy_accepted = models.BooleanField(verbose_name=_("Privacy notice accepted"))
    delete_after_phase = models.BooleanField(
        default=False, verbose_name=_("Delete evaluation data after this phase?")
    )
    privacy_accepted_at = models.DateTimeField(verbose_name=_("Privacy notice accepted at"))
    privacy_form = models.FileField(blank=True, verbose_name=_("Submitted privacy form as PDF"))
    keys = models.ForeignKey(
        to=EvaluationKeyPair,
        on_delete=models.CASCADE,
        verbose_name=_("Keys used to encrypt the evaluation"),
        editable=False,
    )
    finished = models.BooleanField(default=False, verbose_name=_("Finished"))
    finished_at = models.DateTimeField(blank=True, null=True, verbose_name=_("Finished at"))

    deleted = models.BooleanField(default=False, verbose_name=_("Data deleted"))
    deleted_at = models.DateTimeField(blank=True, null=True, verbose_name=_("Data deleted at"))

    objects = EvaluationRegistrationManager.from_queryset(EvaluationRegistrationQuerySet)()

    class Meta:
        verbose_name = _("Evaluation registration")
        verbose_name_plural = _("Evaluation registrations")
        constraints = [
            models.UniqueConstraint(fields=["person", "phase"], name="person_phase_unique")
        ]

    def __str__(self):
        return f"{self.phase}: {self.person}"

    @property
    def finishing_possible(self):
        return (
            bool(self.groups_with_done_evaluations)
            and not self.finished
            and self.phase.status == "evaluation"
        )

    @property
    def results_accessible(self):
        return self.finished and not self.deleted

    def generate_privacy_form(self):
        """Generate a privacy form for this registration."""
        from .tasks import generate_privacy_form_task

        if not self.pk:
            self.save()
        generate_privacy_form_task.delay(self.pk)

    def sync_evaluation_groups(self):
        possible_groups = (
            Group.objects.annotate(members_count=Count("members"))
            .filter(
                owners=self.person,
                members_count__gt=0,
                group_type__in=self.phase.evaluation_group_types.all(),
            )
            .on_day(self.phase.evaluation_date_start)
        )
        evaluation_groups = {g.group: g for g in self.groups.all() if g.group}

        objects_to_add = []
        for group in possible_groups:
            if group not in evaluation_groups:
                evaluation_group = EvaluationGroup(
                    registration=self, group=group, group_name=group.name
                )
                objects_to_add.append(evaluation_group)
            else:
                evaluation_group = evaluation_groups[group]
                if evaluation_group.group_name != group.name:
                    evaluation_group.group_name = group.name
                    evaluation_group.save()

        if objects_to_add:
            EvaluationGroup.objects.bulk_create(objects_to_add)

    @property
    def groups_with_done_evaluations(self) -> QuerySet:
        """Return all groups with at least one done evaluation."""
        return (
            self.groups.all()
            .annotate(
                members_count=Count("group__members", distinct=True),
                done_evaluations_count=Count("done_evaluations", distinct=True),
            )
            .filter(
                done_evaluations_count__gte=get_site_preferences()[
                    "evalu__number_of_persons_results"
                ]
            )
        )

    @property
    def evaluation_groups_with_stats(self) -> QuerySet:
        """Return evaluation groups with some statistics annotated."""
        return self.groups.all().annotate(
            members_count=Count("group__members", distinct=True),
            done_evaluations_count=Count("done_evaluations", distinct=True),
        )

    @classmethod
    def register(
        cls, phase: EvaluationPhase, person: Person, password: str, delete_after_phase: bool = False
    ):
        """Register a person for an evaluation phase."""
        keys = EvaluationKeyPair.create(password)
        registration, __ = EvaluationRegistration.objects.update_or_create(
            phase=phase,
            person=person,
            defaults={
                "privacy_accepted": True,
                "privacy_accepted_at": timezone.now(),
                "delete_after_phase": delete_after_phase,
                "keys": keys,
            },
        )
        return registration

    def finish(self):
        self.finished = True
        self.finished_at = timezone.now()
        self.save()
        self.groups.filter(finished=False).update(
            unlocked=False, finished=True, finished_at=self.finished_at
        )

    def notify_finished_evaluation(self):
        if self.groups.all().count() == self.groups.filter(finished=True).count():
            self.finish()

    def delete_data(self):
        """Delete all data of this registration."""
        if self.deleted or not self.delete_after_phase:
            return
        self.groups.all().delete()
        self.deleted = True
        self.deleted_at = timezone.now()
        self.save()


class EvaluationGroup(ExtensibleModel):
    registration = models.ForeignKey(
        to=EvaluationRegistration,
        on_delete=models.CASCADE,
        verbose_name=_("Evaluated person"),
        related_name="groups",
    )
    group = models.ForeignKey(
        to=Group,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        verbose_name=_("Group of evaluating persons"),
    )
    group_name = models.CharField(max_length=255, verbose_name=_("Group name"))
    unlocked = models.BooleanField(default=False, verbose_name=_("Evaluation unlocked"))

    finished = models.BooleanField(default=False, verbose_name=_("Finished"))
    finished_at = models.DateTimeField(blank=True, null=True, verbose_name=_("Finished at"))

    objects = EvaluationGroupManager.from_queryset(EvaluationGroupQuerySet)()

    class Meta:
        verbose_name = _("Evaluation group")
        verbose_name_plural = _("Evaluation groups")
        constraints = [
            models.UniqueConstraint(
                fields=["registration", "group"], name="registration_group_unique"
            )
        ]

    def __str__(self):
        return f"{self.registration}: {self.group_name}"

    def save(self, *args, **kwargs):
        if self.group:
            self.group_name = self.group.name or self.group.short_name
        super().save(*args, **kwargs)

    @property
    def is_unlocked(self):
        return self.unlocked and self.registration.phase.status == "evaluation"

    @property
    def has_done_evaluations(self):
        return self.done_evaluations.exists()

    def lock(self):
        """Lock the group for evaluations."""
        self.unlocked = False
        self.save()

    def unlock(self):
        """Unlock the group for evaluations."""
        self.unlocked = True
        self.save()

    @property
    def has_enough_evaluations(self):
        return (
            self.done_evaluations.all().count()
            >= get_site_preferences()["evalu__number_of_persons_results"]
        )

    @property
    def finishing_possible(self):
        return (
            not self.finished
            and self.registration.phase.status == "evaluation"
            and self.has_enough_evaluations
        )

    @property
    def results_accessible(self):
        return self.finished and not self.registration.deleted and self.has_enough_evaluations

    def finish(self):
        self.unlocked = False
        self.finished = True
        self.finished_at = timezone.now()
        self.save()
        self.registration.notify_finished_evaluation()

    @property
    def possible_comparison_groups(self):
        return self.registration.phase.comparison_groups.filter(
            Q(groups=self.group) | Q(groups__child_groups=self.group)
        ).distinct()

    def get_results_context(self, private_key: RSAPrivateKey):
        results_per_group = {}
        results_per_group["choices"] = {key: value for key, value in Answer.choices}
        results_per_group["average"] = self.get_average_results(private_key)
        results_per_group["comparison"] = self.get_comparison_results()
        results_per_group["frequency"] = self.get_frequency_results(private_key)
        results_per_group["free_text"] = self.get_free_text_results(private_key)

        by_part = OrderedDict()
        for part in EvaluationPart.objects.order_by("order"):
            by_part[part.id] = {
                "part": model_to_dict(part),
                "results": [],
                "average_results": [],
            }
            for item in part.get_items(self):
                # Agreement result
                if item.item_type == QuestionType.AGREEMENT:
                    result = self.get_frequency_result(item, private_key)
                    by_part[part.id]["results"].append(result)

                # Free text result
                elif item.item_type == QuestionType.FREE_TEXT:
                    result = self.get_free_text_result(item, private_key)
                    by_part[part.id]["results"].append(result)

        for result in results_per_group["average"]:
            part_id = result["part"]["id"]
            by_part.setdefault(part_id, {"part": result["part"], "results": []})
            by_part[part_id]["average_results"].append(result)
        results_per_group["results_by_part"] = by_part
        return results_per_group

    def get_results(self) -> List[any]:
        results = EvaluationResult.objects.filter(group=self)
        return results

    def get_average_results(self, private_key: RSAPrivateKey) -> List[dict]:
        """Get evaluation results as average values for this group."""
        results = []
        for item in EvaluationItem.get_agreement_items(self):
            decrypted_values = [
                r.get_result(private_key) for r in self.results.all() if r.item.id == item.id
            ]
            decrypted_values = [v for v in decrypted_values if v is not None]
            if not decrypted_values:
                continue
            average = sum(decrypted_values) / len(decrypted_values)
            results.append(
                {"item": model_to_dict(item), "average": average, "part": model_to_dict(item.part)}
            )
        return results

    def get_comparison_results(self) -> List[dict]:
        """Get evaluation results as comparison values for this group."""
        # Skip if there is no comparison group
        if not self.group or not self.group.extended_data.get("subject_id"):
            return []
        results = []

        # Get comparison groups and prefetch data
        comparison_groups = self.possible_comparison_groups.prefetch_related(
            Prefetch(
                "results",
                queryset=ComparisonResult.objects.filter(
                    subject_id=self.group.extended_data["subject_id"]
                ),
            )
        )

        for comparison_group in comparison_groups:
            comparison_results = comparison_group.results.all()

            # Check number of results
            if not comparison_group.is_valid_to_show(self.group.subject):
                continue

            result = {
                "comparison_group": model_to_dict(comparison_group, exclude=["groups"]),
                "subject": model_to_dict(self.group.subject),
            }

            # Calculate average results for every item
            concrete_results = []
            for item in EvaluationItem.get_agreement_items(self):
                values = [r.get_result() for r in comparison_results if r.item.id == item.id]
                values = [v for v in values if v is not None]
                if not values:
                    continue
                average = sum(values) / len(values)
                concrete_results.append(
                    {
                        "item": model_to_dict(item),
                        "average": average,
                        "part": model_to_dict(item.part),
                    }
                )
            result["results"] = concrete_results

            results.append(result)
        return results

    def get_frequency_result(
        self, item: "EvaluationItem", private_key: RSAPrivateKey
    ) -> list[dict]:
        labels = {key: desc for key, desc in Answer.choices}
        frequencies = {key: 0 for key, desc in Answer.choices}
        decrypted_values = [
            r.get_result(private_key) for r in self.results.all() if r.item.id == item.id
        ]
        decrypted_values = [v for v in decrypted_values if v is not None]

        for value in decrypted_values:
            frequencies[value] += 1

        frequency_result = OrderedDict()
        for key, value in frequencies.items():
            frequency_result[key] = {
                "label": labels[key],
                "frequency": value,
                "background_color": Answer.background_colors[key],
                "border_color": Answer.border_colors[key],
                "part": model_to_dict(item.part),
            }
        return {
            "item": model_to_dict(item),
            "frequencies": frequency_result,
            "part": model_to_dict(item.part),
            "type": QuestionType.AGREEMENT,
        }

    def get_frequency_results(self, private_key: RSAPrivateKey) -> List[dict]:
        """Get evaluation results as frequencies of single values for this group."""
        results = []
        for item in EvaluationItem.get_agreement_items(self):
            frequency_result = self.get_frequency_result(item, private_key)

            results.append(frequency_result)
        return results

    def get_free_text_result(self, item: "EvaluationItem", private_key: RSAPrivateKey) -> list[str]:
        decrypted_values = [
            r.get_result(private_key) for r in self.results.all() if r.item.id == item.id
        ]
        answers = [v for v in decrypted_values if v]
        return {
            "item": model_to_dict(item),
            "answers": answers,
            "part": model_to_dict(item.part),
            "type": QuestionType.FREE_TEXT,
        }

    def get_free_text_results(self, private_key: RSAPrivateKey) -> List[dict]:
        """Get all free text evaluation results for this group."""
        results = []
        for item in EvaluationItem.get_free_text_items(self):
            answers = self.get_free_text_result(item, private_key)
            results.append(answers)
        return results


class DoneEvaluation(ExtensibleModel):
    group = models.ForeignKey(
        to=EvaluationGroup,
        on_delete=models.CASCADE,
        verbose_name=_("Evaluation group"),
        related_name="done_evaluations",
    )
    evaluated_by = models.ForeignKey(
        to=Person, on_delete=models.CASCADE, verbose_name=_("Evaluated by")
    )

    class Meta:
        verbose_name = _("Done evaluation")
        verbose_name_plural = _("Done evaluations")
        constraints = [
            models.UniqueConstraint(
                fields=["group", "evaluated_by"],
                name="registration_group_evaluated_unique",
            )
        ]

    def __str__(self):
        return f"{self.group}, {self.evaluated_by}"


class EvaluationPart(ExtensibleModel):
    name = models.CharField(max_length=255, verbose_name=_("Name"), unique=True)
    order = models.IntegerField(verbose_name=_("Order"), unique=True)
    optional = models.BooleanField(
        verbose_name=_("Optional part"),
        help_text=_("Students won't have to answer this part."),
        default=False,
    )

    class Meta:
        verbose_name = _("Evaluation part")
        verbose_name_plural = _("Evaluation parts")

    def __str__(self):
        return self.name

    def get_items(self, group: Optional[EvaluationGroup] = None) -> QuerySet["EvaluationItem"]:
        return EvaluationItem.get_items(group).filter(part=self)


class QuestionType(models.TextChoices):
    FREE_TEXT = "free_text", _("Free text")
    AGREEMENT = "agreement", _("Agreement")


class Answer(models.IntegerChoices):
    TRUE = 1, _("Is true")
    MOSTLY_TRUE = 2, _("Is mostly true")
    PARTIALLY_TRUE = 3, _("Is partially true")
    LESS_TRUE = 4, _("Is less true")
    NOT_TRUE = 5, _("Is not true")

    @classproperty
    def background_colors(cls) -> Dict[int, str]:  # noqa
        return {
            1: "#88bf4a",
            2: "#bbd154",
            3: "#efe258",
            4: "#f29e58",
            5: "#eb595a",
        }

    @classproperty
    def border_colors(cls) -> Dict[int, str]:  # noqa
        return {
            1: "#88bf4a",
            2: "#bbd154",
            3: "#efe258",
            4: "#f29e58",
            5: "#eb595a",
        }


class EvaluationItem(ExtensiblePolymorphicModel):
    part = models.ForeignKey(
        to=EvaluationPart,
        on_delete=models.CASCADE,
        verbose_name=_("Evaluation part"),
        related_name="items",
    )
    order = models.IntegerField(verbose_name=_("Order"))
    name = models.CharField(
        max_length=255, verbose_name=_("Name"), help_text=_("as shown in results")
    )
    question = models.TextField(
        verbose_name=_("Question"), help_text=_("as shown on evaluation form")
    )
    item_type = models.CharField(
        choices=QuestionType.choices, max_length=255, verbose_name=_("Question type")
    )

    class Meta:
        verbose_name = _("Evaluation item")
        verbose_name_plural = _("Evaluation items")

    def __str__(self):
        return self.name

    @classmethod
    def get_items(cls, group: Optional[EvaluationGroup] = None) -> QuerySet["EvaluationItem"]:
        qs1 = cls.objects.not_instance_of(CustomEvaluationItem)
        if group:
            qs2 = CustomEvaluationItem.objects.filter(group=group)
            qs1 = cls.objects.filter(Q(pk__in=qs1) | Q(pk__in=qs2))
        qs1 = qs1.order_by("part__order", "-polymorphic_ctype__model", "order")
        return qs1

    @classmethod
    def get_agreement_items(
        cls, group: Optional[EvaluationGroup] = None
    ) -> QuerySet["EvaluationItem"]:
        return cls.get_items(group).filter(item_type=QuestionType.AGREEMENT)

    @classmethod
    def get_free_text_items(
        cls, group: Optional[EvaluationGroup] = None
    ) -> QuerySet["EvaluationItem"]:
        return cls.get_items(group).filter(item_type=QuestionType.FREE_TEXT)


class CustomEvaluationItem(EvaluationItem):
    group = models.ForeignKey(
        to=EvaluationGroup,
        on_delete=models.CASCADE,
        verbose_name=_("Evaluation group"),
        related_name="custom_items",
    )

    class Meta:
        verbose_name = _("Custom Evaluation Item")
        verbose_name_plural = _("Custom Evaluation Items")

    def __str__(self):
        return super().__str__()


class EvaluationResult(ExtensibleModel):
    group = models.ForeignKey(
        to=EvaluationGroup,
        on_delete=models.CASCADE,
        verbose_name=_("Evaluation group"),
        related_name="results",
    )
    item = models.ForeignKey(
        to=EvaluationItem,
        on_delete=models.CASCADE,
        verbose_name=_("Question"),
        related_name="results",
    )
    result = models.TextField(verbose_name=_("Encrypted result"))
    result_key = models.TextField(verbose_name=_("Encrypted fernet key"))

    class Meta:
        verbose_name = _("Evaluation result")
        verbose_name_plural = _("Evaluation results")

    def __str__(self):
        return f"{self.group}, {self.item}"

    def store_result(self, result: Union[str, int], commit: bool = False):
        if not self.group:
            raise ValueError("No evaluation group set: encryption impossible.")
        message = str(result)

        fernet_key = Fernet.generate_key()

        # Encrypt fernet key and store it
        encrypted_fernet_key = self.group.registration.keys.encrypt(fernet_key)
        self.result_key = encrypted_fernet_key

        # Encrypt message and store it
        fernet = Fernet(fernet_key)
        self.result = b64encode(fernet.encrypt(message.encode())).decode()

        if commit:
            self.save()

    def add_comparison_results(self, result: str):
        if not self.group.group or not self.group.group.extended_data.get("subject_id"):
            # Can't store comparison results without a group
            return
        subject = self.group.group.extended_data.get("subject_id")
        for comparison_group in self.group.possible_comparison_groups:
            if comparison_group.is_valid_to_store(subject):
                comparison_result = ComparisonResult(
                    comparison_group=comparison_group,
                    subject=subject,
                    item=self.item,
                    result=result,
                )
                comparison_result.save()
                DoneEvaluationComparison.objects.get_or_create(
                    comparison_group=comparison_group, evaluation_group=self.group
                )

    def get_result(self, private_key: RSAPrivateKey) -> Union[str, int]:
        # Decrypt fernet key
        fernet_key = self.group.registration.keys.decrypt(self.result_key, private_key)

        # Decrypt result
        fernet = Fernet(fernet_key)
        result = fernet.decrypt(b64decode(self.result.encode())).decode()

        if self.item.item_type == QuestionType.AGREEMENT:
            return int(result) if result else None
        return result


class ComparisonResult(ExtensibleModel):
    comparison_group = models.ForeignKey(
        to=ComparisonGroup,
        on_delete=models.CASCADE,
        verbose_name=_("Comparison group"),
        related_name="results",
    )
    subject = models.ForeignKey(
        to=Subject,
        on_delete=models.CASCADE,
        verbose_name=_("Subject"),
        related_name="evalu_results",
    )

    item = models.ForeignKey(
        to=EvaluationItem,
        on_delete=models.CASCADE,
        verbose_name=_("Question"),
        related_name="comparison_results",
    )
    result = models.TextField(verbose_name=_("Result"))

    class Meta:
        verbose_name = _("Comparison result")
        verbose_name_plural = _("Comparison results")

    def __str__(self):
        return f"{self.comparison_group}, {self.subject}, {self.item}"

    def get_result(self) -> Union[str, int]:
        if self.item.item_type == QuestionType.AGREEMENT:
            return int(self.result) if self.result else None
        return self.result


class DoneEvaluationComparison(ExtensibleModel):
    comparison_group = models.ForeignKey(
        to=ComparisonGroup,
        on_delete=models.CASCADE,
        verbose_name=_("Comparison group"),
        related_name="done_evaluations_comparison",
    )
    evaluation_group = models.ForeignKey(
        to=EvaluationGroup,
        on_delete=models.CASCADE,
        verbose_name=_("Evaluation group"),
        related_name="done_evaluations_comparison",
    )

    class Meta:
        verbose_name = _("Done evaluation (comparison)")
        verbose_name_plural = _("Done evaluations (comparison)")

    def __str__(self):
        return f"{self.comparison_group}, {self.evaluation_group}"



# Generated by Django 3.2.8 on 2021-10-17 15:18

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('evalu', '0005_key_model'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='EvaluationKeySet',
            new_name='EvaluationKeyPair',
        ),
        migrations.AddField(
            model_name='evaluationresult',
            name='group',
            field=models.ForeignKey(default=0, on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationgroup', verbose_name='Evaluation group'),
            preserve_default=False,
        ),
    ]



# Generated by Django 3.2.8 on 2021-11-07 17:24

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0020_pdf_file_person_optional'),
        ('evalu', '0006_key_pair'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='doneevaluation',
            name='registration_group_evaluated_by_site_unique',
        ),
        migrations.RemoveField(
            model_name='doneevaluation',
            name='registration',
        ),
        migrations.RemoveField(
            model_name='evaluationresult',
            name='registration',
        ),
        migrations.AlterField(
            model_name='doneevaluation',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='done_evaluations', to='evalu.evaluationgroup', verbose_name='Evaluation group'),
        ),
        migrations.AlterField(
            model_name='evaluationgroup',
            name='registration',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='groups', to='evalu.evaluationregistration', verbose_name='Evaluated person'),
        ),
        migrations.AlterField(
            model_name='evaluationitem',
            name='part',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items', to='evalu.evaluationpart', verbose_name='Evaluation part'),
        ),
        migrations.AlterField(
            model_name='evaluationregistration',
            name='person',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='evaluation_registrations', to='core.person', verbose_name='Evaluated person'),
        ),
        migrations.AlterField(
            model_name='evaluationresult',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='evalu.evaluationgroup', verbose_name='Evaluation group'),
        ),
        migrations.AddConstraint(
            model_name='doneevaluation',
            constraint=models.UniqueConstraint(fields=('group', 'evaluated_by'),
                                               name='group_evaluated_by_site_unique'),
        ),
    ]



from aleksis.core.util.apps import AppConfig


class DefaultConfig(AppConfig):
    name = "aleksis.apps.evalu"
    verbose_name = "EvaLU (Evaluation of teaching and lesson quality)"
    dist_name = "AlekSIS-App-EvaLU"

    urls = {
        "Repository": "https://edugit.org/katharineum/AlekSIS-App-EvaLU",
    }
    licence = "EUPL-1.2+"
    copyright_info = (([2021, 2022, 2023], "Jonathan Weth", "dev@jonathanweth.de"),)



from datetime import date, timedelta

from django.utils import timezone

from aleksis.apps.evalu.models import (
    EvaluationItem,
    EvaluationKeyPair,
    EvaluationPart,
    EvaluationPhase,
    EvaluationRegistration,
    QuestionType,
)
from aleksis.core.models import Group, Person, SchoolTerm


def _get_test_person():
    return Person.objects.get_or_create(first_name="Jane", last_name="Doe")[0]


def _get_test_group():
    return Group.objects.get_or_create(name="Test group")[0]


def _get_test_phase():
    return EvaluationPhase.objects.get_or_create(
        name="Test Phase",
        defaults=dict(
            evaluated_group=_get_test_group(),
            privacy_notice="Some dummy content",
            registration_date_start=date(2020, 1, 1),
            registration_date_end=date(2020, 1, 2),
            evaluation_date_start=date(2020, 1, 3),
            evaluation_date_end=date(2020, 1, 4),
        ),
    )[0]


def _get_test_evaluation():
    return EvaluationRegistration.objects.create(
        phase=_get_test_phase(),
        person=_get_test_person(),
        privacy_accepted=True,
        privacy_accepted_at=timezone.now(),
        keys=EvaluationKeyPair.create("test"),
    )


def _get_test_school_term():
    school_term, __ = SchoolTerm.objects.get_or_create(
        name="Test", date_start=date(2019, 8, 1), date_end=date(2020, 6, 30)
    )
    return school_term


def _create_some_persons(prefix, n):
    persons = []
    for i in range(n):
        p = Person.objects.create(
            first_name=f"{prefix} First name {i}", last_name=f"{prefix} Last name {i}"
        )
        persons.append(p)
    return persons


def _create_some_empty_groups():
    school_term = _get_test_school_term()
    groups = []
    for i in range(5):
        g = Group.objects.create(name=f"Empty Group {i}", school_term=school_term)
        g.owners.add(_get_test_person())
        groups.append(g)
    return groups


def _create_some_full_groups():
    school_term = _get_test_school_term()
    groups = []
    for i in range(5):
        persons = _create_some_persons("Full Group", 10)
        g = Group.objects.create(name=f"Full Group {i}", school_term=school_term)
        g.members.set(persons)
        g.owners.add(_get_test_person())
        groups.append(g)
    return groups


def _get_test_part():
    return EvaluationPart.objects.get_or_create(name="Foo", defaults=dict(order=1))[0]


def _get_test_item():
    return EvaluationItem.objects.get_or_create(
        part=_get_test_part(),
        defaults=dict(name="Foo", question="Foo?", order=1, item_type=QuestionType.AGREEMENT),
    )[0]



from django import forms
from django.core.exceptions import ValidationError
from django.forms import RadioSelect
from django.utils.translation import gettext as _

from django_select2.forms import ModelSelect2MultipleWidget, ModelSelect2Widget
from material import Fieldset, Layout, Row, Span2, Span3, Span5, Span6, Span8

from aleksis.apps.evalu.models import (
    ComparisonGroup,
    CustomEvaluationItem,
    EvaluationGroup,
    EvaluationItem,
    EvaluationPart,
    EvaluationPhase,
)


class EvaluationPartForm(forms.ModelForm):
    class Meta:
        model = EvaluationPart
        fields = ["name", "order", "optional"]


class EvaluationItemForm(forms.ModelForm):
    layout = Layout(
        Row(Span2("order"), Span3("item_type"), Span5("name"), Span2("DELETE")), Row("question")
    )

    class Meta:
        model = EvaluationItem
        fields = ["order", "name", "question", "item_type"]


EvaluationItemFormSet = forms.inlineformset_factory(
    EvaluationPart, EvaluationItem, form=EvaluationItemForm, min_num=1
)


class CustomEvaluationItemForm(forms.ModelForm):
    layout = Layout(
        Row(Span8("part"), Span2("order"), Span2("DELETE")),
        Row(
            Span6("item_type"),
            Span6("name"),
        ),
        Row("question"),
    )

    class Meta:
        model = CustomEvaluationItem
        fields = ["part", "order", "name", "question", "item_type"]


CustomEvaluationItemFormSet = forms.inlineformset_factory(
    EvaluationGroup, CustomEvaluationItem, form=CustomEvaluationItemForm, min_num=0
)


class EvaluationPhaseForm(forms.ModelForm):
    layout = Layout(
        Fieldset(
            _("Base data"), Row("name"), Row("evaluated_group"), Row("evaluation_group_types")
        ),
        Fieldset(
            _("Time range"),
            Row("registration_date_start", "registration_date_end"),
            Row("evaluation_date_start", "evaluation_date_end"),
        ),
        Fieldset(_("Additional data"), Row("privacy_notice")),
    )

    class Meta:
        model = EvaluationPhase
        fields = [
            "name",
            "evaluated_group",
            "evaluation_group_types",
            "registration_date_start",
            "registration_date_end",
            "evaluation_date_start",
            "evaluation_date_end",
            "privacy_notice",
        ]
        widgets = {
            "evaluated_group": ModelSelect2Widget(
                search_fields=["name__icontains", "short_name__icontains"],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            )
        }


class ComparisonGroupForm(forms.ModelForm):
    layout = Layout(
        Row(Span5("name"), Span5("groups"), Span2("DELETE")),
    )

    class Meta:
        model = ComparisonGroup
        fields = ["name", "groups"]
        widgets = {
            "groups": ModelSelect2MultipleWidget(
                search_fields=["name__icontains", "short_name__icontains"],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            )
        }


ComparisonGroupFormSet = forms.inlineformset_factory(
    EvaluationPhase, ComparisonGroup, form=ComparisonGroupForm, min_num=0
)


class RegisterForEvaluationForm(forms.Form):
    layout = Layout(Row("consent"), Row("delete_after_phase"))
    consent = forms.BooleanField(
        required=True,
        label=_(
            "I hereby declare that I agree that my data may be collected "
            "and processed in the above sense within the context of EvaLU."
        ),
        widget=forms.CheckboxInput(attrs={"class": "filled-in"}),
    )
    delete_after_phase = forms.BooleanField(
        initial=False,
        required=False,
        label=_("My data should be deleted after each evaluation phase."),
        widget=forms.CheckboxInput(attrs={"class": "filled-in"}),
    )
    password = forms.CharField(
        required=True, label=_("Evaluation password"), widget=forms.PasswordInput()
    )
    password_confirm = forms.CharField(
        required=True, label=_("Confirm evaluation password"), widget=forms.PasswordInput()
    )

    def clean(self):
        cleaned_data = super().clean()

        if cleaned_data["password"] != cleaned_data["password_confirm"]:
            raise ValidationError("You have entered two different passwords.")


class AgreementWidget(RadioSelect):
    pass


class EvaluationFinishForm(forms.Form):
    pass


class PasswordForm(forms.Form):
    password = forms.CharField(
        required=True, label=_("Evaluation password"), widget=forms.PasswordInput()
    )



from django.utils import timezone

import pytest

from aleksis.apps.evalu.models import EvaluationGroup, EvaluationKeyPair, EvaluationRegistration
from aleksis.apps.evalu.tests.models.util import (
    _create_some_empty_groups,
    _create_some_full_groups,
    _create_some_persons,
    _get_test_evaluation,
    _get_test_person,
    _get_test_phase,
    _get_test_school_term,
)
from aleksis.core.models import Group

pytestmark = pytest.mark.django_db


def test_create():
    registration = EvaluationRegistration.objects.create(
        phase=_get_test_phase(),
        person=_get_test_person(),
        privacy_accepted=True,
        privacy_accepted_at=timezone.now(),
        keys=EvaluationKeyPair.create("test"),
    )
    assert registration.pk is not None


def test_str():
    registration = _get_test_evaluation()
    assert "Doe" in str(registration)
    assert "Test Phase" in str(registration)


def test_group_sync():
    empty_groups = _create_some_empty_groups()
    full_groups = _create_some_full_groups()
    registration = _get_test_evaluation()

    assert EvaluationGroup.objects.all().count() == 0
    registration.sync_evaluation_groups()
    assert EvaluationGroup.objects.all().count() == len(full_groups)

    for group in EvaluationGroup.objects.all():
        assert group.registration == registration
        assert group.group
        assert "Full" in group.group_name
        assert group.group in full_groups
        assert group.group not in empty_groups


def _prepare_evaluation_groups():
    empty_groups = _create_some_empty_groups()
    full_groups = _create_some_full_groups()
    registration = _get_test_evaluation()
    registration.sync_evaluation_groups()
    return full_groups, empty_groups, registration


def test_group_sync_rename_one():
    full_groups, empty_groups, registration = _prepare_evaluation_groups()
    person = _get_test_person()

    # Get test group
    one_group = Group.objects.get(name="Full Group 1")

    assert EvaluationGroup.objects.get(group=one_group).group_name == "Full Group 1"

    # Change group and sync
    one_group.name = "Renamed Group"
    one_group.save()
    registration.sync_evaluation_groups()

    assert EvaluationGroup.objects.get(group=one_group).group_name == "Renamed Group"


def test_group_sync_add_one():
    full_groups, empty_groups, registration = _prepare_evaluation_groups()
    person = _get_test_person()

    # Create a new group
    new_group = Group.objects.create(name="New Group", school_term=_get_test_school_term())
    persons = _create_some_persons("New Group", 10)
    new_group.owners.add(person)
    new_group.members.set(persons)

    assert EvaluationGroup.objects.all().count() == len(full_groups)
    registration.sync_evaluation_groups()
    assert EvaluationGroup.objects.all().count() == len(full_groups) + 1


def test_group_sync_delete_one():
    full_groups, empty_groups, registration = _prepare_evaluation_groups()

    # Get test group
    one_group = Group.objects.get(name="Full Group 1")

    assert EvaluationGroup.objects.all().count() == len(full_groups)
    group_count = Group.objects.all().count()

    # Delete groups and sync
    one_group.delete()
    registration.sync_evaluation_groups()

    assert EvaluationGroup.objects.all().count() == len(full_groups)
    assert Group.objects.all().count() == group_count - 1



# Generated by Django 5.1.1 on 2024-10-06 12:24

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cursus', '0003_drop_site'),
        ('evalu', '0021_evaluationgroup_finished_evaluationgroup_finished_at'),
    ]

    operations = [
        migrations.AlterField(
            model_name='comparisonresult',
            name='subject',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='evalu_results', to='cursus.subject', verbose_name='Subject'),
        ),
    ]



from django.db.models import Prefetch, Q, QuerySet

from aleksis.core.managers import AlekSISBaseManagerWithoutMigrations
from aleksis.core.models import Person


class EvaluationPhaseManager(AlekSISBaseManagerWithoutMigrations):
    """Manager adding specific methods to evaluation phases."""

    def get_queryset(self):
        """Ensure all related data is loaded as well."""
        return super().get_queryset().select_related("evaluated_group")


class EvaluationPhaseQuerySet(QuerySet):
    """Custom QuerySet for evaluation phases."""

    def for_person(self, person: Person) -> "EvaluationPhaseQuerySet":
        return self.filter(evaluated_group__members=person)

    def not_registered(self, person: Person) -> "EvaluationPhaseQuerySet":
        return self.exclude(registrations__person=person)

    def can_register(self, person: Person) -> "EvaluationPhaseQuerySet":
        return self.for_person(person).not_registered(person)

    def for_person_with_registrations(self, person: Person) -> "EvaluationPhaseQuerySet":
        from aleksis.apps.evalu.models import EvaluationRegistration

        return self.for_person(person).prefetch_related(
            Prefetch(
                "registrations",
                queryset=EvaluationRegistration.objects.filter(person=person),
            )
        )


class EvaluationRegistrationManager(AlekSISBaseManagerWithoutMigrations):
    """Manager adding specific methods to evaluation registrations."""

    def get_queryset(self):
        """Ensure all related data is loaded as well."""
        return super().get_queryset().select_related("phase", "person", "keys")


class EvaluationRegistrationQuerySet(QuerySet):
    """Custom QuerySet for evaluation registrations."""


class EvaluationGroupManager(AlekSISBaseManagerWithoutMigrations):
    """Manager adding specific methods to evaluation groups."""

    def get_queryset(self):
        """Ensure all related data is loaded as well."""
        return super().get_queryset().select_related("registration", "group")


class EvaluationGroupQuerySet(QuerySet):
    """Custom QuerySet for evaluation groups."""

    def for_person(self, person: Person) -> "EvaluationGroupQuerySet":
        """Get all groups for which a person can evaluate."""
        return self.filter(Q(group__members=person) | Q(done_evaluations__evaluated_by=person))

    def for_person_with_done_evaluations(self, person: Person) -> "EvaluationGroupQuerySet":
        from aleksis.apps.evalu.models import DoneEvaluation

        return self.for_person(person).prefetch_related(
            Prefetch(
                "done_evaluations",
                queryset=DoneEvaluation.objects.filter(evaluated_by=person),
            )
        )



import pytest

from aleksis.apps.evalu.models import EvaluationKeyPair

pytestmark = pytest.mark.django_db


def test_create():
    pair = EvaluationKeyPair.create("test")
    assert pair.pk is not None


def test_encryption():
    pair1 = EvaluationKeyPair.create("test")
    pair2 = EvaluationKeyPair.create("test2")

    assert pair1.encrypt("foo")
    assert pair1.encrypt("foo") != pair1.encrypt("foo")
    assert pair1.encrypt("foo") != pair1.encrypt("baz")
    assert pair1.encrypt("foo") != pair2.encrypt("foo")


def test_decryption():
    pair1 = EvaluationKeyPair.create("test")
    pair2 = EvaluationKeyPair.create("test2")

    assert pair1.decrypt(pair1.encrypt("foo"), "test").decode() == "foo"
    assert pair1.decrypt(pair1.encrypt("baz"), "test").decode() == "baz"

    with pytest.raises(ValueError):
        pair1.decrypt(pair1.encrypt("foo"), "test2")

    with pytest.raises(ValueError):
        pair1.decrypt(pair2.encrypt("foo"), "test")


def test_unlock():
    pair1 = EvaluationKeyPair.create("test")

    assert pair1.test("test")

    with pytest.raises(ValueError):
        pair1.test("test2")


def test_str():
    pair1 = EvaluationKeyPair.create("test")
    assert "Key" in str(pair1)


def test_broken():
    pair1 = EvaluationKeyPair()
    assert pair1.get_public_key() is None
    assert pair1.get_private_key("foo") is None



from django.contrib import admin

from .models import (
    DoneEvaluation,
    EvaluationGroup,
    EvaluationItem,
    EvaluationKeyPair,
    EvaluationPart,
    EvaluationPhase,
    EvaluationRegistration,
    EvaluationResult,
)

admin.site.register(EvaluationRegistration)
admin.site.register(EvaluationGroup)
admin.site.register(EvaluationResult)
admin.site.register(EvaluationItem)
admin.site.register(EvaluationPart)
admin.site.register(EvaluationPhase)
admin.site.register(DoneEvaluation)
admin.site.register(EvaluationKeyPair)



from rules import add_perm

from aleksis.apps.evalu.util.predicates import (
    are_results_accessible,
    has_any_evaluation_group,
    has_done_evaluations,
    is_evaluated_person,
    is_evaluated_person_for_group,
    is_evaluation_status,
    is_evaluation_status_for_group,
    is_finished,
    is_finishing_possible,
    is_member_of_any_evaluation_group,
    is_member_of_evaluated_group,
    is_participant_for_group,
    is_registration_running,
    is_unlocked,
)
from aleksis.core.util.predicates import has_global_perm, has_object_perm, has_person

view_evaluationparts_predicate = has_person & has_global_perm("evalu.view_evaluationpart")
add_perm("evalu.view_evaluationparts_rule", view_evaluationparts_predicate)

add_evaluationpart_predicate = view_evaluationparts_predicate & has_global_perm(
    "evalu.add_evaluationpart"
)
add_perm("evalu.add_evaluationpart_rule", add_evaluationpart_predicate)

edit_evaluationpart_predicate = view_evaluationparts_predicate & has_global_perm(
    "evalu.change_evaluationpart"
)
add_perm("evalu.edit_evaluationpart_rule", edit_evaluationpart_predicate)

delete_evaluationpart_predicate = view_evaluationparts_predicate & has_global_perm(
    "evalu.delete_evaluationpart"
)
add_perm("evalu.delete_evaluationpart_rule", delete_evaluationpart_predicate)

view_evaluationphases_predicate = has_person & has_global_perm("evalu.view_evaluationphase")
add_perm("evalu.view_evaluationphases_rule", view_evaluationphases_predicate)

view_evaluationphase_predicate = has_person & (
    has_global_perm("evalu.view_evaluationphase") | has_object_perm("evalu.view_evaluationphase")
)
add_perm("evalu.view_evaluationphase_rule", view_evaluationphase_predicate)

create_evaluationphase_predicate = view_evaluationphases_predicate & has_global_perm(
    "evalu.create_evaluationphase"
)
add_perm("evalu.create_evaluationphase_rule", create_evaluationphase_predicate)

edit_evaluationphase_predicate = view_evaluationphase_predicate & (
    has_global_perm("evalu.change_evaluationphase")
    | has_object_perm("evalu.change_evaluationphase")
)
add_perm("evalu.edit_evaluationphase_rule", edit_evaluationphase_predicate)


delete_data_predicate = (
    view_evaluationphase_predicate
    & (
        has_global_perm("evalu.change_evaluationphase")
        | has_object_perm("evalu.change_evaluationphase")
    )
    & is_evaluation_status("results")
)
add_perm("evalu.delete_data_rule", delete_data_predicate)

delete_evaluationphase_predicate = view_evaluationphase_predicate & (
    has_global_perm("evalu.delete_evaluationphase")
    | has_object_perm("evalu.delete_evaluationphase")
)
add_perm("evalu.delete_evaluationphase_rule", delete_evaluationphase_predicate)

view_evaluationphases_overview_predicate = has_person & (is_member_of_any_evaluation_group)
add_perm("evalu.view_evaluationphases_overview_rule", view_evaluationphases_overview_predicate)

register_for_evaluation_predicate = (
    has_person & is_member_of_evaluated_group & is_registration_running
)
add_perm("evalu.register_for_evaluation_rule", register_for_evaluation_predicate)

view_evaluationregistration_predicate = has_person & (
    is_evaluated_person | has_global_perm("evalu.view_evaluationregistration")
)
add_perm("evalu.view_evaluationregistration_rule", view_evaluationregistration_predicate)

manage_evaluation_process_predicate = (
    view_evaluationregistration_predicate & is_evaluation_status("evaluation") & ~is_finished
)
add_perm("evalu.manage_evaluation_process_rule", manage_evaluation_process_predicate)

view_evaluations_results_predicate = view_evaluationregistration_predicate & (
    is_evaluation_status("results") | is_finished
)
add_perm("evalu.view_evaluation_results_rule", view_evaluations_results_predicate)

view_evaluations_results_for_group_predicate = (
    (is_evaluated_person_for_group | has_global_perm("evalu.view_evaluationgroup"))
    & (is_evaluation_status_for_group("results") | is_finished)
    & are_results_accessible
)
add_perm(
    "evalu.view_evaluation_results_for_group_rule", view_evaluations_results_for_group_predicate
)

finish_evaluation_predicate = view_evaluationregistration_predicate & is_finishing_possible
add_perm("evalu.finish_evaluation_rule", finish_evaluation_predicate)

edit_custom_evaluation_items_for_group_predicate = (
    has_person
    & ~is_finished
    & ~has_done_evaluations
    & (is_evaluated_person_for_group | has_global_perm("evalu.edit_evaluationgroup"))
)
add_perm(
    "evalu.edit_custom_evaulation_items_for_group_rule",
    edit_custom_evaluation_items_for_group_predicate,
)

start_evaluation_for_group_predicate = (
    has_person
    & is_evaluation_status_for_group("evaluation")
    & ~is_finished
    & ~is_unlocked
    & (is_evaluated_person_for_group | has_global_perm("evalu.edit_evaluationgroup"))
)
add_perm("evalu.start_evaluation_for_group_rule", start_evaluation_for_group_predicate)

stop_evaluation_for_group_predicate = (
    has_person
    & is_evaluation_status_for_group("evaluation")
    & ~is_finished
    & is_unlocked
    & (is_evaluated_person_for_group | has_global_perm("evalu.edit_evaluationgroup"))
)
add_perm("evalu.stop_evaluation_for_group_rule", stop_evaluation_for_group_predicate)

finish_evaluation_for_group_predicate = (
    is_evaluated_person_for_group | has_global_perm("evalu.edit_evaluationgroup")
) & is_finishing_possible
add_perm("evalu.finish_evaluation_for_group_rule", finish_evaluation_for_group_predicate)


view_evaluations_as_participant_predicate = has_person & has_any_evaluation_group
add_perm("evalu.view_evaluations_as_participant_rule", view_evaluations_as_participant_predicate)

evaluate_person_predicate = (
    has_person
    & is_participant_for_group
    & is_evaluation_status_for_group("evaluation")
    & is_unlocked
)
add_perm("evalu.evaluate_person_rule", evaluate_person_predicate)

view_evaluation_menu_predicate = has_person & (
    view_evaluationparts_predicate
    | view_evaluationphases_predicate
    | view_evaluationphases_overview_predicate
    | view_evaluations_as_participant_predicate
)
add_perm("evalu.view_menu_rule", view_evaluation_menu_predicate)



# Generated by Django 3.2.16 on 2022-12-26 10:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('evalu', '0011_evaluation_finished'),
    ]

    operations = [
        migrations.AddField(
            model_name='evaluationregistration',
            name='deleted',
            field=models.BooleanField(default=False, verbose_name='Data deleted'),
        ),
        migrations.AddField(
            model_name='evaluationregistration',
            name='deleted_at',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Data deleted at'),
        ),
    ]



# Generated by Django 3.2.6 on 2021-08-25 17:40

import ckeditor.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0020_pdf_file_person_optional'),
    ]

    operations = [
        migrations.CreateModel(
            name='EvaluationPart',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('order', models.IntegerField(verbose_name='Order')),
            ],
            options={
                'verbose_name': 'Evaluation part',
                'verbose_name_plural': 'Evaluation parts',
            },
        ),
        migrations.CreateModel(
            name='EvaluationPhase',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('name', models.CharField(max_length=255, verbose_name='Display Name')),
                ('registration_date_start', models.DateField(verbose_name='First date teachers can register themselves for evaluation')),
                ('registration_date_end', models.DateField(verbose_name='Last date teachers can register themselves for evaluation')),
                ('evaluation_date_start', models.DateField(verbose_name='First date teachers can start the evaluation')),
                ('evaluation_date_end', models.DateField(verbose_name='Date when all evaluations stop')),
                ('results_date_start', models.DateField(verbose_name='First date teachers can see results')),
                ('privacy_notice', ckeditor.fields.RichTextField(verbose_name='Privacy notice which teachers have to agree')),
                ('evaluated_group', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.group', verbose_name='Group with teachers which can register for evaluation')),
            ],
            options={
                'verbose_name': 'Evaluation phase',
                'verbose_name_plural': 'Evaluation phases',
            },
        ),
        migrations.CreateModel(
            name='EvaluationQuestion',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('order', models.IntegerField(verbose_name='Order')),
                ('name', models.CharField(help_text='as shown in results', max_length=255, verbose_name='Name')),
                ('question', ckeditor.fields.RichTextField(help_text='as shown on evaluation form', verbose_name='Question')),
                ('question_type', models.CharField(choices=[('free_text', 'Free text'), ('agreement', 'Agreement')], max_length=255, verbose_name='Question type')),
                ('part', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationpart', verbose_name='Evaluation part')),
            ],
            options={
                'verbose_name': 'Evaluation question',
                'verbose_name_plural': 'Evaluation questions',
            },
        ),
        migrations.CreateModel(
            name='EvaluationRegistration',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('privacy_accepted', models.BooleanField(verbose_name='Privacy notice accepted')),
                ('privacy_accepted_at', models.DateTimeField(verbose_name='Privacy notice accepted at')),
                ('gpg_keys', models.TextField(editable=False, verbose_name='GPG keys used to encrypt the evaluation')),
                ('person', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.person', verbose_name='Evaluated person')),
                ('phase', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationphase', verbose_name='Evaluation phase')),
            ],
            options={
                'verbose_name': 'Evaluation registration',
                'verbose_name_plural': 'Evaluation registrations',
            },
        ),
        migrations.CreateModel(
            name='EvaluationResult',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('result', models.TextField(verbose_name='Encrypted result')),
                ('question', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationquestion', verbose_name='Question')),
                ('registration', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationregistration', verbose_name='Evaluated person')),
            ],
            options={
                'verbose_name': 'Evaluation result',
                'verbose_name_plural': 'Evaluation results',
            },
        ),
        migrations.CreateModel(
            name='EvaluationGroup',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('unlocked', models.BooleanField(default=False, verbose_name='Evaluation unlocked')),
                ('group', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.group', verbose_name='Group of evaluating persons')),
                ('registration', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationregistration', verbose_name='Evaluated person')),
            ],
            options={
                'verbose_name': 'Evaluation group',
                'verbose_name_plural': 'Evaluation groups',
            },
        ),
        migrations.CreateModel(
            name='DoneEvaluation',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('evaluated_by', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.person', verbose_name='Evaluated by')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationgroup', verbose_name='Evaluation group')),
                ('registration', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationregistration', verbose_name='Evaluated person')),
            ],
            options={
                'verbose_name': 'Done evaluation',
                'verbose_name_plural': 'Done evaluations',
            },
        ),
        migrations.AddConstraint(
            model_name='evaluationregistration',
            constraint=models.UniqueConstraint(fields=('person', 'phase',), name='person_phase_site_unique'),
        ),
        migrations.AddConstraint(
            model_name='evaluationquestion',
            constraint=models.UniqueConstraint(fields=('part', 'order'), name='part_site_order_unique'),
        ),
        migrations.AddConstraint(
            model_name='evaluationpart',
            constraint=models.UniqueConstraint(fields=('order',), name='order_site_unique'),
        ),
        migrations.AddConstraint(
            model_name='evaluationpart',
            constraint=models.UniqueConstraint(fields=('name',), name='name_site_unique'),
        ),
        migrations.AddConstraint(
            model_name='evaluationgroup',
            constraint=models.UniqueConstraint(fields=('registration', 'group',), name='registration_group_site_unique'),
        ),
        migrations.AddConstraint(
            model_name='doneevaluation',
            constraint=models.UniqueConstraint(fields=('registration', 'group', 'evaluated_by',), name='registration_group_evaluated_by_site_unique'),
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-23 18:56

from django.db import migrations, models
import django.db.models.deletion


def forwards_func(apps, schema_editor):
    # We get the model from the versioned app registry;
    # if we directly import it, it'll be the wrong version
    ContentType = apps.get_model("contenttypes", "ContentType")
    EvaluationItem = apps.get_model("evalu", "EvaluationItem")
    db_alias = schema_editor.connection.alias
    ct = ContentType.objects.using(db_alias).get(app_label="evalu", model="evaluationitem")
    EvaluationItem.objects.using(db_alias).update(
        polymorphic_ctype_id=ct.id
    )

class Migration(migrations.Migration):

    dependencies = [
        ("evalu", "0014_evaluationitem_polymorphic"),
    ]

    operations = [
        migrations.RunPython(
            forwards_func
        ),
    ]



import pytest

from aleksis.apps.evalu.models import (
    EvaluationGroup,
    EvaluationPart,
    EvaluationResult,
    QuestionType,
)
from aleksis.apps.evalu.tests.models.util import _get_test_evaluation, _get_test_item

pytestmark = pytest.mark.django_db


def test_store_int():
    registration = _get_test_evaluation()
    item = _get_test_item()
    group = EvaluationGroup.objects.create(
        registration=registration,
        group_name="Test Group",
    )

    result = EvaluationResult(group=group, item=item)
    assert result.result == ""

    result.store_result(10)
    assert result.result != ""
    assert result.result != "10"

    assert result.get_result("test") == 10


def test_store_str():
    registration = _get_test_evaluation()
    item = _get_test_item()
    item.item_type = QuestionType.FREE_TEXT
    item.save()
    group = EvaluationGroup.objects.create(
        registration=registration,
        group_name="Test Group",
    )

    result = EvaluationResult(group=group, item=item)
    assert result.result == ""

    result.store_result("I am a string")
    assert result.result != ""
    assert result.result != "I am a string"

    assert result.get_result("test") == "I am a string"


LONG_STR = """
Lorem ipsum dolor sit amet, consetetur sadipscing elitr, 
sed diam nonumy eirmod tempor invidunt ut labore et dolore 
magna aliquyam erat, sed diam voluptua. At vero eos et accusam 
et justo duo dolores et ea rebum. Stet clita kasd gubergren, 
no sea takimata sanctus est Lorem ipsum dolor sit amet. 
Lorem ipsum dolor sit amet, consetetur sadipscing elitr, 
sed diam nonumy eirmod tempor invidunt ut labore et dolore 
magna aliquyam erat, sed diam voluptua. 
At vero eos et accusam et justo duo dolores et ea rebum. 
Stet clita kasd gubergren, 
no sea takimata sanctus est Lorem ipsum dolor sit amet.
"""


def test_store_too_long_str():
    registration = _get_test_evaluation()
    item = _get_test_item()
    item.item_type = QuestionType.FREE_TEXT
    item.save()
    group = EvaluationGroup.objects.create(
        registration=registration,
        group_name="Test Group",
    )

    result = EvaluationResult(group=group, item=item)
    assert result.result == ""

    result.store_result(LONG_STR)

    assert result.result != ""
    assert result.result != LONG_STR

    assert result.get_result("test") == LONG_STR



# Generated by Django 3.2.7 on 2021-10-08 19:11

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('evalu', '0003_question_no_rich_text'),
    ]

    operations = [
        migrations.AddField(
            model_name='evaluationregistration',
            name='delete_after_phase',
            field=models.BooleanField(default=False, verbose_name='Delete evaluation data after this phase?'),
        ),
        migrations.AddField(
            model_name='evaluationregistration',
            name='privacy_form',
            field=models.FileField(blank=True, null=True, upload_to='', verbose_name='Submitted privacy form as PDF'),
        ),
        migrations.AlterField(
            model_name='evaluationregistration',
            name='phase',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='registrations', to='evalu.evaluationphase', verbose_name='Evaluation phase'),
        ),
    ]



from django.db.models import Q
from django.utils import timezone

from rules import predicate

from aleksis.apps.evalu.models import EvaluationGroup, EvaluationPhase


@predicate
def is_member_of_any_evaluation_group(user):
    return EvaluationPhase.objects.filter(evaluated_group__members=user.person).exists()


@predicate
def is_member_of_evaluated_group(user, obj):
    return user.person in obj.evaluated_group.members.all()


@predicate
def is_evaluated_person(user, obj):
    return user.person == obj.person


@predicate
def is_registration_running(user, obj):
    if hasattr(obj, "phase"):
        return obj.phase.registration_running
    return obj.registration_running


def is_evaluation_status(status):
    name = f"is_evaluation_status:{status}"

    @predicate(name)
    def fn(user, obj):
        if hasattr(obj, "phase"):
            return obj.phase.status == status
        return obj.status == status

    return fn


def is_evaluation_status_for_group(status):
    name = f"is_evaluation_status_for_group:{status}"

    @predicate(name)
    def fn(user, obj):
        return obj.registration.phase.status == status

    return fn


@predicate
def is_evaluated_person_for_group(user, obj):
    return user.person == obj.registration.person


@predicate
def is_participant_for_group(user, obj):
    if not obj.group:
        return False
    return user.person in obj.group.members.all()


@predicate
def is_unlocked(user, obj):
    return obj.is_unlocked


@predicate
def has_any_evaluation_group(user, obj):
    now_date = timezone.now().date()
    return (
        EvaluationGroup.objects.filter(
            Q(group__members=user.person) | Q(done_evaluations__evaluated_by=user.person)
        )
        .filter(registration__phase__evaluation_date_start__lte=now_date)
        .exists()
    )


@predicate
def is_finishing_possible(user, obj):
    return obj.finishing_possible


@predicate
def is_finished(user, obj):
    return obj.finished


@predicate
def has_done_evaluations(user, obj: EvaluationGroup):
    return obj.has_done_evaluations


@predicate
def are_results_accessible(user, obj: EvaluationGroup):
    return obj.results_accessible



from django.template.loader import render_to_string
from django.utils.translation import gettext as _

import django_tables2 as tables


class EvaluationPartTable(tables.Table):
    """Table to list evaluation parts."""

    class Meta:
        attrs = {"class": "responsive-table highlight"}
        order_by = "order"

    name = tables.LinkColumn("edit_evaluation_part", args=[tables.A("pk")])
    edit = tables.LinkColumn(
        "edit_evaluation_part",
        args=[tables.A("id")],
        text=_("Edit"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange orange-text"}},
    )
    delete = tables.LinkColumn(
        "delete_evaluation_part",
        args=[tables.A("id")],
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red red-text"}},
    )


class EvaluationPhaseTable(tables.Table):
    """Table to list evaluation phases."""

    class Meta:
        attrs = {"class": "responsive-table highlight"}

    status = tables.Column()
    name = tables.LinkColumn("evaluation_phase", args=[tables.A("id")])

    edit = tables.LinkColumn(
        "edit_evaluation_phase",
        args=[tables.A("id")],
        text=_("Edit"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange orange-text"}},
    )
    delete = tables.LinkColumn(
        "delete_evaluation_phase",
        args=[tables.A("id")],
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red red-text"}},
    )

    def render_status(self, value, record):
        return render_to_string("evalu/phase/status.html", dict(phase=record))



# Generated by Django 4.2.6 on 2023-10-09 14:12

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("evalu", "0012_deletion"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="evaluationphase",
            name="results_date_start",
        ),
    ]



from datetime import date

from django.core.exceptions import ValidationError

import pytest
from freezegun import freeze_time

from aleksis.apps.evalu.tests.models.util import _get_test_phase

pytestmark = pytest.mark.django_db


def test_create():
    phase = _get_test_phase()
    phase.save()

    assert phase.pk is not None


def test_clean_registration_date_start_before_end():
    phase = _get_test_phase()

    phase.clean()  # Correct result

    phase.registration_date_start = date(2020, 1, 2)
    phase.registration_date_end = date(2020, 1, 1)
    with pytest.raises(ValidationError):
        phase.clean()


def test_clean_evaluation_date_start_before_registration_date_end():
    phase = _get_test_phase()

    phase.registration_date_start = date(2020, 1, 1)
    phase.registration_date_end = date(2020, 1, 2)
    phase.evaluation_date_start = date(2020, 1, 2)
    phase.evaluation_date_end = date(2020, 1, 3)
    phase.clean()

    phase.evaluation_date_start = date(2020, 1, 1)
    with pytest.raises(ValidationError):
        phase.clean()

    phase.evaluation_date_start = date(2020, 1, 3)
    phase.clean()


def test_clean_evaluation_date_start_before_date_end():
    phase = _get_test_phase()

    phase.evaluation_date_start = date(2020, 1, 3)
    phase.evaluation_date_end = date(2020, 1, 4)
    with pytest.raises(ValidationError):
        phase.clean()


def test_evaluation_status():
    phase = _get_test_phase()
    phase.registration_date_start = date(2020, 1, 1)
    phase.registration_date_end = date(2020, 1, 2)
    phase.evaluation_date_start = date(2020, 1, 4)
    phase.evaluation_date_end = date(2020, 1, 5)
    phase.results_date_start = date(2020, 1, 7)

    with freeze_time("2019-12-31"):
        assert phase.status == "not_started"
    with freeze_time("2020-01-01"):
        assert phase.status == "not_started"
        assert phase.registration_running
    with freeze_time("2020-01-02"):
        assert phase.status == "not_started"
        assert phase.registration_running
    with freeze_time("2020-01-03"):
        assert phase.status == "not_started"
        assert not phase.registration_running
    with freeze_time("2020-01-04"):
        assert phase.status == "evaluation"
    with freeze_time("2020-01-05"):
        assert phase.status == "evaluation"
    with freeze_time("2020-01-06"):
        assert phase.status == "evaluation_closed"
    with freeze_time("2020-01-07"):
        assert phase.status == "evaluation_closed"


def test_str():
    phase = _get_test_phase()
    assert str(phase) == "Test Phase"



# Generated by Django 3.2.11 on 2022-01-22 16:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('evalu', '0010_evaluationresult_result_key'),
    ]

    operations = [
        migrations.AddField(
            model_name='evaluationregistration',
            name='finished',
            field=models.BooleanField(default=False, verbose_name='Finished'),
        ),
        migrations.AddField(
            model_name='evaluationregistration',
            name='finished_at',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Finished at'),
        ),
    ]



# Generated by Django 4.2.6 on 2023-11-03 19:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("evalu", "0017_remove_evaluationitem_part_site_order_unique"),
    ]

    operations = [
        migrations.AddField(
            model_name="evaluationpart",
            name="optional",
            field=models.BooleanField(
                default=False,
                help_text="Students won't have to answer this part.",
                verbose_name="Optional part",
            ),
        ),
    ]



# Generated by Django 3.2.9 on 2022-01-04 12:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('evalu', '0009_auto_20211206_1448'),
    ]

    operations = [
        migrations.AddField(
            model_name='evaluationresult',
            name='result_key',
            field=models.TextField(default='', verbose_name='Encrypted fernet key'),
            preserve_default=False,
        ),
    ]



# Generated by Django 3.2.9 on 2021-12-05 18:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('chronos', '0008_unique_constraints'),
        ('core', '0020_pdf_file_person_optional'),
        ('evalu', '0007_drop_reference_registration'),
    ]

    operations = [
        migrations.CreateModel(
            name='ComparisonGroup',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
            ],
            options={
                'verbose_name': 'Comparison group',
                'verbose_name_plural': 'Comparison groups',
            },
        ),
        migrations.CreateModel(
            name='ComparisonResult',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('result', models.TextField(verbose_name='Result')),
            ],
            options={
                'verbose_name': 'Comparison result',
                'verbose_name_plural': 'Comparison results',
            },
        ),
        migrations.AlterModelManagers(
            name='evaluationphase',
            managers=[
            ],
        ),
        migrations.AlterModelManagers(
            name='evaluationregistration',
            managers=[
            ],
        ),
        migrations.RemoveConstraint(
            model_name='doneevaluation',
            name='group_evaluated_by_site_unique',
        ),
        migrations.AlterField(
            model_name='evaluationregistration',
            name='privacy_form',
            field=models.FileField(blank=True, default='', upload_to='', verbose_name='Submitted privacy form as PDF'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='evaluationresult',
            name='item',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='evalu.evaluationitem', verbose_name='Question'),
        ),
        migrations.AddConstraint(
            model_name='doneevaluation',
            constraint=models.UniqueConstraint(fields=('group', 'evaluated_by'), name='registration_group_evaluated_by_site_unique'),
        ),
        migrations.AddField(
            model_name='comparisonresult',
            name='comparison_group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='evalu.comparisongroup', verbose_name='Comparison group'),
        ),
        migrations.AddField(
            model_name='comparisonresult',
            name='item',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='comparison_results', to='evalu.evaluationitem', verbose_name='Question'),
        ),
        migrations.AddField(
            model_name='comparisonresult',
            name='subject',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='evalu_results', to='chronos.subject', verbose_name='Subject'),
        ),
        migrations.AddField(
            model_name='comparisongroup',
            name='groups',
            field=models.ManyToManyField(to='core.Group', verbose_name='Groups'),
        ),
        migrations.AddField(
            model_name='comparisongroup',
            name='phase',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='comparison_groups', to='evalu.evaluationphase', verbose_name='Evaluation phase'),
        ),
    ]



from django.urls import path

from . import views

urlpatterns = [
    path("parts/", views.EvaluationPartListView.as_view(), name="evaluation_parts"),
    path("parts/create/", views.EvaluationPartCreateView.as_view(), name="create_evaluation_part"),
    path("parts/<int:pk>/", views.EvaluationPartEditView.as_view(), name="edit_evaluation_part"),
    path(
        "parts/<int:pk>/delete/",
        views.EvaluationPartDeleteView.as_view(),
        name="delete_evaluation_part",
    ),
    path("phases/", views.EvaluationPhaseListView.as_view(), name="evaluation_phases"),
    path(
        "phases/create/", views.EvaluationPhaseCreateView.as_view(), name="create_evaluation_phase"
    ),
    path("phases/<int:pk>/", views.EvaluationPhaseDetailView.as_view(), name="evaluation_phase"),
    path(
        "phases/<int:pk>/edit/",
        views.EvaluationPhaseEditView.as_view(),
        name="edit_evaluation_phase",
    ),
    path(
        "phases/<int:pk>/delete/",
        views.EvaluationPhaseDeleteView.as_view(),
        name="delete_evaluation_phase",
    ),
    path(
        "phases/<int:pk>/deletion/",
        views.DeleteDataView.as_view(),
        name="delete_data_from_phase",
    ),
    path(
        "evaluations/",
        views.EvaluationPhaseOverviewView.as_view(),
        name="evaluation_phases_overview",
    ),
    path(
        "evaluations/<int:pk>/register/",
        views.RegisterForEvaluationView.as_view(),
        name="register_for_evaluation",
    ),
    path(
        "evaluations/registrations/<int:pk>/",
        views.RegistrationDetailView.as_view(),
        name="evaluation_registration",
    ),
    path(
        "evaluations/registrations/<int:pk>/manage/",
        views.ManageEvaluationProcessView.as_view(),
        name="manage_evaluation_process",
    ),
    path(
        "evaluations/groups/<int:pk>/start/",
        views.StartEvaluationForGroupView.as_view(),
        name="start_evaluation_for_group",
    ),
    path(
        "evaluations/groups/<int:pk>/stop/",
        views.StopEvaluationForGroupView.as_view(),
        name="stop_evaluation_for_group",
    ),
    path(
        "evaluations/groups/<int:pk>/finish/",
        views.FinishEvaluationForGroupView.as_view(),
        name="finish_evaluation_for_group",
    ),
    path(
        "evaluations/groups/<int:pk>/custom_items/",
        views.CustomEvaluationItemsEditView.as_view(),
        name="edit_custom_evaluation_items",
    ),
    path(
        "evaluations/registrations/<int:pk>/finish/",
        views.FinishEvaluationView.as_view(),
        name="finish_evaluation",
    ),
    path(
        "evaluations/registrations/<int:pk>/results/",
        views.EvaluationResultsView.as_view(),
        {"as_pdf": False},
        name="evaluation_results",
    ),
    path(
        "evaluations/registrations/<int:pk>/results/pdf/",
        views.EvaluationResultsView.as_view(),
        {"as_pdf": True},
        name="evaluation_results_as_pdf",
    ),
    path(
        "evaluations/groups/<int:pk>/results/",
        views.EvaluationResultsForGroupView.as_view(),
        {"as_pdf": False},
        name="evaluation_results_for_group",
    ),
    path(
        "evaluations/groups/<int:pk>/results/pdf/",
        views.EvaluationResultsForGroupView.as_view(),
        {"as_pdf": True},
        name="evaluation_results_for_group_as_pdf",
    ),
    path(
        "evaluations/evaluate/",
        views.EvaluationsAsParticipantListView.as_view(),
        name="evaluations_as_participant",
    ),
    path(
        "evaluations/evaluate/<int:pk>/",
        views.EvaluationFormView.as_view(),
        name="evaluate_person",
    ),
]



# Generated by Django 3.2.7 on 2021-10-12 13:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('evalu', '0004_privacy_form'),
    ]

    operations = [
        migrations.CreateModel(
            name='EvaluationKeySet',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('private_key', models.TextField(editable=False, verbose_name='Private key')),
                ('public_key', models.TextField(editable=False, verbose_name='Public key')),
            ],
            options={
                'verbose_name': 'Evaluation key set',
                'verbose_name_plural': 'Evaluation key set',
            },
        ),
        migrations.RemoveField(
            model_name='evaluationregistration',
            name='gpg_keys',
        ),
        migrations.AddField(
            model_name='evaluationregistration',
            name='keys',
            field=models.ForeignKey(default=0, editable=False, on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationkeyset', verbose_name='Keys used to encrypt the evaluation'),
            preserve_default=False,
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-23 19:03

import aleksis.core.managers
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("evalu", "0015_migrate_evaluationitem_poly"),
    ]

    operations = [
        migrations.CreateModel(
            name="CustomEvaluationItem",
            fields=[
                (
                    "evaluationitem_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="evalu.evaluationitem",
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="custom_items",
                        to="evalu.evaluationgroup",
                        verbose_name="Evaluation group",
                    ),
                ),
            ],
            options={
                "verbose_name": "Custom Evaluation Item",
                "verbose_name_plural": "Custom Evaluation Items",
            },
            bases=("evalu.evaluationitem",),
        ),
    ]



from collections import OrderedDict
from typing import Any, Dict, Optional

from django.contrib import messages
from django.core.exceptions import ValidationError
from django.forms import CharField, Form, Textarea, TypedChoiceField, model_to_dict
from django.shortcuts import redirect, render
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.views import View
from django.views.decorators.cache import never_cache
from django.views.generic import DetailView, ListView
from django.views.generic.detail import SingleObjectMixin

from django_tables2 import SingleTableView
from formtools.wizard.views import CookieWizardView
from material import Layout
from reversion.views import RevisionMixin
from rules.contrib.views import PermissionRequiredMixin

from aleksis.apps.evalu.forms import (
    AgreementWidget,
    ComparisonGroupFormSet,
    CustomEvaluationItemFormSet,
    EvaluationFinishForm,
    EvaluationItemFormSet,
    EvaluationPartForm,
    EvaluationPhaseForm,
    PasswordForm,
    RegisterForEvaluationForm,
)
from aleksis.apps.evalu.models import (
    Answer,
    DoneEvaluation,
    EvaluationGroup,
    EvaluationPart,
    EvaluationPhase,
    EvaluationRegistration,
    EvaluationResult,
    QuestionType,
)
from aleksis.apps.evalu.tables import EvaluationPartTable, EvaluationPhaseTable
from aleksis.core.mixins import AdvancedCreateView, AdvancedDeleteView, AdvancedEditView
from aleksis.core.util.pdf import render_pdf


class EvaluationPartListView(PermissionRequiredMixin, SingleTableView):
    """Table of all extra marks."""

    model = EvaluationPart
    table_class = EvaluationPartTable
    permission_required = "evalu.view_evaluationparts_rule"
    template_name = "evalu/part/list.html"


@method_decorator(never_cache, name="dispatch")
class EvaluationPartCreateView(PermissionRequiredMixin, AdvancedCreateView):
    """Create view for extra marks."""

    model = EvaluationPart
    form_class = EvaluationPartForm
    permission_required = "evalu.add_evaluationpart_rule"
    template_name = "evalu/part/create.html"
    success_message = _("The evaluation part has been created.")

    def get_success_url(self):
        return reverse("edit_evaluation_part", args=[self.object.pk])


@method_decorator(never_cache, name="dispatch")
class EvaluationPartDeleteView(PermissionRequiredMixin, RevisionMixin, AdvancedDeleteView):
    """Delete view for extra marks."""

    model = EvaluationPart
    permission_required = "evalu.delete_evaluationpart_rule"
    template_name = "core/pages/delete.html"
    success_url = reverse_lazy("evaluation_parts")
    success_message = _("The evaluation part has been deleted.")


@method_decorator(never_cache, name="dispatch")
class EvaluationPartEditView(PermissionRequiredMixin, AdvancedEditView):
    model = EvaluationPart
    form_class = EvaluationPartForm
    success_message = _("The evaluation part and it's items have been updated successfully.")
    permission_required = "evalu.edit_evaluationpart_rule"
    template_name = "evalu/part/edit.html"

    def get_success_url(self):
        return reverse("edit_evaluation_part", args=[self.object.pk])

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        self.formset = EvaluationItemFormSet(self.request.POST or None, instance=self.object)
        context["formset"] = self.formset
        return context

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.get_context_data(**kwargs)
        form = self.get_form()
        if form.is_valid() and self.formset.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def form_valid(self, form):
        self.object = form.save()
        self.formset.instance = self.object
        self.formset.save()
        return super().form_valid(form)


class EvaluationPhaseListView(PermissionRequiredMixin, SingleTableView):
    """View to list all evaluation phases."""

    model = EvaluationPhase
    table_class = EvaluationPhaseTable
    permission_required = "evalu.view_evaluationphases_rule"
    template_name = "evalu/phase/list.html"


class EvaluationPhaseDetailView(PermissionRequiredMixin, DetailView):
    """Detail view for evaluation phases."""

    model = EvaluationPhase
    permission_required = "evalu.view_evaluationphase_rule"
    template_name = "evalu/phase/detail.html"

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context["members"] = self.object.members_with_registration
        return context


@method_decorator(never_cache, name="dispatch")
class EvaluationPhaseCreateView(PermissionRequiredMixin, AdvancedCreateView):
    """Create view for evaluation phases."""

    model = EvaluationPhase
    form_class = EvaluationPhaseForm
    permission_required = "evalu.create_evaluationphase_rule"
    template_name = "evalu/phase/create.html"
    success_message = _("The evaluation phase has been created.")

    def get_success_url(self):
        return reverse("edit_evaluation_phase", args=[self.object.pk])


@method_decorator(never_cache, name="dispatch")
class EvaluationPhaseDeleteView(PermissionRequiredMixin, RevisionMixin, AdvancedDeleteView):
    """Delete view for evaluation phases."""

    model = EvaluationPhase
    permission_required = "evalu.delete_evaluationphase_rule"
    template_name = "core/pages/delete.html"
    success_url = reverse_lazy("evaluation_phases")
    success_message = _("The evaluation phase has been deleted.")


@method_decorator(never_cache, name="dispatch")
class EvaluationPhaseEditView(PermissionRequiredMixin, AdvancedEditView):
    model = EvaluationPhase
    form_class = EvaluationPhaseForm
    success_message = _("The evaluation phase has been updated.")
    permission_required = "evalu.edit_evaluationphase_rule"
    template_name = "evalu/phase/edit.html"

    def get_success_url(self):
        return reverse("evaluation_phase", args=[self.object.pk])

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        self.formset = ComparisonGroupFormSet(self.request.POST or None, instance=self.object)
        context["formset"] = self.formset
        return context

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.get_context_data(**kwargs)
        form = self.get_form()
        if form.is_valid() and self.formset.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def form_valid(self, form):
        self.object = form.save()
        self.formset.instance = self.object
        self.formset.save()
        return super().form_valid(form)


class EvaluationPhaseOverviewView(PermissionRequiredMixin, ListView):
    """View to list all evaluation phases a user can register or is registered for."""

    model = EvaluationPhase
    permission_required = "evalu.view_evaluationphases_overview_rule"
    template_name = "evalu/list.html"

    def get_queryset(self):
        return EvaluationPhase.objects.for_person_with_registrations(self.request.user.person)


class DeleteDataView(PermissionRequiredMixin, DetailView):
    """View to delete data after a certain date."""

    model = EvaluationPhase
    permission_required = "evalu.delete_data_rule"
    template_name = "evalu/phase/delete_data.html"

    def post(self, request, *args, **kwargs):
        if request.POST.get("delete"):
            self.object = self.get_object()
            self.object.delete_data()
            messages.success(request, _("The data have been deleted."))
            return redirect("evaluation_phase", pk=self.object.pk)
        else:
            return self.get(request, *args, **kwargs)


@method_decorator(never_cache, name="dispatch")
class RegisterForEvaluationView(PermissionRequiredMixin, DetailView):
    """View to register for an evaluation phase."""

    model = EvaluationPhase
    permission_required = "evalu.register_for_evaluation_rule"
    template_name = "evalu/registration/register.html"

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context["form"] = RegisterForEvaluationForm(self.request.POST or None)
        return context

    def get_queryset(self):
        return EvaluationPhase.objects.can_register(self.request.user.person)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)

        form = context["form"]

        if form.is_valid():
            data = form.cleaned_data
            registration = EvaluationRegistration.register(
                self.object, self.request.user.person, data["password"], data["delete_after_phase"]
            )
            registration.generate_privacy_form()
            messages.success(
                request, _("You have successfully registered yourself for the evaluation.")
            )
            return redirect("evaluation_registration", registration.pk)

        return self.render_to_response(context)


class EvaluationGroupMixin:
    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        self.object.sync_evaluation_groups()
        context["possible_groups"] = self.object.evaluation_groups_with_stats

        return context


class RegistrationDetailView(PermissionRequiredMixin, EvaluationGroupMixin, DetailView):
    model = EvaluationRegistration
    permission_required = "evalu.view_evaluationregistration_rule"
    template_name = "evalu/registration/detail.html"


@method_decorator(never_cache, name="dispatch")
class ManageEvaluationProcessView(PermissionRequiredMixin, EvaluationGroupMixin, DetailView):
    model = EvaluationRegistration
    permission_required = "evalu.manage_evaluation_process_rule"
    template_name = "evalu/registration/manage.html"


@method_decorator(never_cache, name="dispatch")
class StartEvaluationForGroupView(PermissionRequiredMixin, SingleObjectMixin, View):
    permission_required = "evalu.start_evaluation_for_group_rule"
    model = EvaluationGroup

    def get(self, request, *args, **kwargs):
        group = self.get_object()
        group.unlock()
        messages.success(
            request,
            _("The evaluation for the group {} has been successfully unlocked.").format(
                group.group_name
            ),
        )
        return redirect("manage_evaluation_process", group.registration.pk)


@method_decorator(never_cache, name="dispatch")
class StopEvaluationForGroupView(PermissionRequiredMixin, SingleObjectMixin, View):
    permission_required = "evalu.stop_evaluation_for_group_rule"
    model = EvaluationGroup

    def get(self, request, *args, **kwargs):
        group = self.get_object()
        group.lock()
        messages.success(
            request,
            _("The evaluation for the group {} has been successfully locked.").format(
                group.group_name
            ),
        )
        return redirect("manage_evaluation_process", group.registration.pk)


@method_decorator(never_cache, name="dispatch")
class FinishEvaluationForGroupView(PermissionRequiredMixin, DetailView):
    model = EvaluationGroup
    permission_required = "evalu.finish_evaluation_for_group_rule"
    template_name = "evalu/registration/finish_group.html"

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()

        if "finish" in request.POST:
            self.object.finish()
            return redirect("manage_evaluation_process", self.object.registration.pk)

        return self.get(request, *args, **kwargs)


class EvaluationsAsParticipantListView(PermissionRequiredMixin, ListView):
    permission_required = "evalu.view_evaluations_as_participant_rule"
    template_name = "evalu/participate/list.html"

    def get_queryset(self):
        now_date = timezone.now().date()
        person = self.request.user.person
        return (
            EvaluationGroup.objects.for_person_with_done_evaluations(person)
            .filter(registration__phase__evaluation_date_start__lte=now_date)
            .distinct()
        )


@method_decorator(never_cache, name="dispatch")
class EvaluationFormView(PermissionRequiredMixin, SingleObjectMixin, CookieWizardView):
    permission_required = "evalu.evaluate_person_rule"
    model = EvaluationGroup
    template_name = "evalu/participate/form.html"
    form_list = [EvaluationFinishForm]

    def __init__(self, *args, **kwargs):
        self.field_mapping = {}
        self.parts = {}
        super().__init__(*args, **kwargs)

    def get_form_list(self):
        form_list = self._build_forms(self.object)
        computed_form_list = OrderedDict()
        for i, form in enumerate(form_list):
            computed_form_list[str(i)] = form

        return computed_form_list

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["parts"] = self.parts

        if self.steps.current != self.steps.last:
            context["part"] = self.parts[self.steps.index]

        # Get data for preview table (last step)
        if self.steps.current == self.steps.last:
            forms = []
            for i, form_key in enumerate(self.get_form_list()):
                form = self.get_form(
                    step=form_key,
                    data=self.storage.get_step_data(form_key),
                    files=self.storage.get_step_files(form_key),
                )

                # Skip the form if it has no fields (e. g. finish form)
                if not form.fields:
                    continue

                # Make sure cleaned_data is filled
                form.is_valid()

                form_data = []
                for key, value in form.fields.items():
                    value_to_show = form.cleaned_data.get(key)

                    # Use speaking value from choices if set
                    if getattr(value, "choices", None):
                        value_to_show = dict(value.choices).get(value_to_show, "–")

                    form_data.append((value.label, value_to_show))
                forms.append((self.parts[i], form_data))
            context["form_data"] = forms

        return context

    def _build_forms(self, group: Optional[EvaluationGroup] = None):
        forms = []

        for i, part in enumerate(EvaluationPart.objects.order_by("order")):
            self.parts[i] = part
            layout = []
            form_class_attrs = {}
            agreement_keys = []

            for item in part.get_items(group):
                field_name = f"item_{item.pk}"

                # Build different types of form fields
                if item.item_type == QuestionType.AGREEMENT:
                    field = TypedChoiceField(
                        coerce=int,
                        choices=Answer.choices,
                        label=item.question,
                        widget=AgreementWidget(),
                        required=not part.optional,
                    )
                    agreement_keys.append(field_name)
                elif item.item_type == QuestionType.FREE_TEXT:
                    field = CharField(
                        required=False, label=item.question, widget=Textarea(attrs={"rows": 4})
                    )
                else:
                    continue

                form_class_attrs[field_name] = field
                self.field_mapping[field_name] = item
                layout.append(field_name)

            def clean_method(self, part=part, agreement_keys=agreement_keys):  # noqa
                if part.optional:
                    agreement_data = [
                        self.cleaned_data.get(key)
                        for key in agreement_keys
                        if self.cleaned_data.get(key)
                    ]
                    if len(agreement_data) != len(agreement_keys) and len(agreement_data) != 0:
                        raise ValidationError(
                            _("You have to answer no or all questions in an optional part.")
                        )

            # Build final evaluation form
            form_class_attrs["layout"] = Layout(*layout)
            form = type(Form)(f"EvaluationForm{part.pk}", (Form,), form_class_attrs)
            form.clean = clean_method

            forms.append(form)

        # Add dummy form for review page (just to make sure that it appears)
        forms.append(EvaluationFinishForm)
        self.parts[len(forms) - 1] = {"name": _("Finish")}

        return forms

    def done(self, form_list, **kwargs):
        # Prevent errors with doubled evaluations
        if DoneEvaluation.objects.filter(
            group=self.object, evaluated_by=self.request.user.person
        ).exists():
            return redirect("evaluations_as_participant")

        done_evaluation = DoneEvaluation(group=self.object, evaluated_by=self.request.user.person)
        for form in form_list:
            for key, value in form.cleaned_data.items():
                item = self.field_mapping[key]
                result = EvaluationResult(group=self.object, item=item)
                result.store_result(value)
                result.save()
                result.add_comparison_results(value)
        done_evaluation.save()
        messages.success(
            self.request,
            _("The evaluation has been finished. Thank you for doing this evaluation!"),
        )
        return redirect("evaluations_as_participant")


@method_decorator(never_cache, name="dispatch")
class FinishEvaluationView(PermissionRequiredMixin, DetailView):
    model = EvaluationRegistration
    permission_required = "evalu.finish_evaluation_rule"
    template_name = "evalu/registration/finish.html"

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()

        if "finish" in request.POST:
            self.object.finish()
            return redirect("evaluation_phases_overview")

        return self.get(request, *args, **kwargs)


@method_decorator(never_cache, name="dispatch")
class EvaluationResultsView(PermissionRequiredMixin, EvaluationGroupMixin, DetailView):
    model = EvaluationRegistration
    permission_required = "evalu.view_evaluation_results_rule"
    template_name = "evalu/password.html"

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)

        context["as_pdf"] = self.kwargs.get("as_pdf", False)

        self.form = PasswordForm(self.request.POST or None)
        context["form"] = self.form

        return context

    def post(self, request, as_pdf: bool, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)

        if self.form.is_valid():
            password = self.form.cleaned_data["password"]
            keys = self.object.keys

            try:
                private_key = keys.get_private_key(password)
            except ValueError:
                messages.error(
                    request,
                    _(
                        "There was an error with decrypting the data. "
                        "Please check if you have entered the correct password."
                    ),
                )
                return self.render_to_response(context)
        else:
            return self.render_to_response(context)

        # Result calculation
        groups = self.object.groups_with_done_evaluations.prefetch_related("results")
        context["groups"] = groups

        all_results = []
        for group in groups:
            all_results.append((model_to_dict(group), group.get_results_context(private_key)))

        context["results"] = all_results

        if as_pdf:
            return render_pdf(request, "evalu/results/all_pdf.html", context)

        return render(request, "evalu/results/all.html", context)


@method_decorator(never_cache, name="dispatch")
class EvaluationResultsForGroupView(PermissionRequiredMixin, DetailView):
    model = EvaluationGroup
    permission_required = "evalu.view_evaluation_results_for_group_rule"
    template_name = "evalu/password.html"

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)

        context["as_pdf"] = self.kwargs.get("as_pdf", False)

        self.form = PasswordForm(self.request.POST or None)
        context["form"] = self.form

        return context

    def post(self, request, as_pdf: bool, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)

        if self.form.is_valid():
            password = self.form.cleaned_data["password"]
            keys = self.object.registration.keys

            try:
                private_key = keys.get_private_key(password)
            except ValueError:
                messages.error(
                    request,
                    _(
                        "There was an error with decrypting the data. "
                        "Please check if you have entered the correct password."
                    ),
                )
                return self.render_to_response(context)
        else:
            return self.render_to_response(context)

        context["group"] = model_to_dict(self.object)
        context["results"] = [
            (model_to_dict(self.object), self.object.get_results_context(private_key))
        ]

        if as_pdf:
            return render_pdf(request, "evalu/results/group_pdf.html", context)

        return render(request, "evalu/results/group.html", context)


@method_decorator(never_cache, name="dispatch")
class CustomEvaluationItemsEditView(PermissionRequiredMixin, DetailView):
    model = EvaluationGroup
    permission_required = "evalu.edit_custom_evaulation_items_for_group_rule"
    template_name = "evalu/registration/edit_custom_items.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        self.formset = CustomEvaluationItemFormSet(self.request.POST or None, instance=self.object)
        context["formset"] = self.formset
        return context

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.get_context_data(**kwargs)
        if self.formset.is_valid():
            self.formset.instance = self.object
            self.formset.save()
            messages.success(
                request, _("The custom evaluation items have been saved successfully.")
            )
        #       else:
        #     return self.form_invalid(form)
        #
        return super().get(request, *args, **kwargs)



# Generated by Django 5.0.8 on 2024-08-24 14:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0065_calendarevent_update_timezone_constraints'),
        ('evalu', '0018_evaluationpart_optional'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='doneevaluation',
            name='registration_group_evaluated_by_site_unique',
        ),
        migrations.RemoveConstraint(
            model_name='evaluationgroup',
            name='registration_group_site_unique',
        ),
        migrations.RemoveConstraint(
            model_name='evaluationpart',
            name='order_site_unique',
        ),
        migrations.RemoveConstraint(
            model_name='evaluationpart',
            name='name_site_unique',
        ),
        migrations.RemoveConstraint(
            model_name='evaluationregistration',
            name='person_phase_site_unique',
        ),
    ] + [
        migrations.RunSQL(
            f"ALTER TABLE evalu_{model_name} drop column if exists site_id;"
        ) for model_name  in
        [
            "evaluationphase",
            "comparisongroup",
            "evaluationregistration",
            "evaluationgroup",
            "doneevaluation",
            "evaluationpart",
            "evaluationitem",
            "customevaluationitem",
            "evaluationresult",
            "comparisonresult",
            "doneevaluationcomparison",
        ]
    ] + [
        migrations.AddField(
            model_name='comparisongroup',
            name='managed_by_app_label',
            field=models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance'),
        ),
        migrations.AddField(
            model_name='comparisonresult',
            name='managed_by_app_label',
            field=models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance'),
        ),
        migrations.AddField(
            model_name='doneevaluation',
            name='managed_by_app_label',
            field=models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance'),
        ),
        migrations.AddField(
            model_name='doneevaluationcomparison',
            name='managed_by_app_label',
            field=models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance'),
        ),
        migrations.AddField(
            model_name='evaluationgroup',
            name='managed_by_app_label',
            field=models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance'),
        ),
        migrations.AddField(
            model_name='evaluationitem',
            name='managed_by_app_label',
            field=models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance'),
        ),
        migrations.AddField(
            model_name='evaluationpart',
            name='managed_by_app_label',
            field=models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance'),
        ),
        migrations.AddField(
            model_name='evaluationphase',
            name='managed_by_app_label',
            field=models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance'),
        ),
        migrations.AddField(
            model_name='evaluationregistration',
            name='managed_by_app_label',
            field=models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance'),
        ),
        migrations.AddField(
            model_name='evaluationresult',
            name='managed_by_app_label',
            field=models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance'),
        ),
        migrations.AlterField(
            model_name='evaluationpart',
            name='name',
            field=models.CharField(max_length=255, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='evaluationpart',
            name='order',
            field=models.IntegerField(unique=True, verbose_name='Order'),
        ),
        migrations.AddConstraint(
            model_name='doneevaluation',
            constraint=models.UniqueConstraint(fields=('group', 'evaluated_by'), name='registration_group_evaluated_unique'),
        ),
        migrations.AddConstraint(
            model_name='evaluationgroup',
            constraint=models.UniqueConstraint(fields=('registration', 'group'), name='registration_group_unique'),
        ),
        migrations.AddConstraint(
            model_name='evaluationregistration',
            constraint=models.UniqueConstraint(fields=('person', 'phase'), name='person_phase_unique'),
        ),
    ]



# Generated by Django 3.2.6 on 2021-09-15 15:08

import ckeditor.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('evalu', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='EvaluationItem',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('order', models.IntegerField(verbose_name='Order')),
                ('name', models.CharField(help_text='as shown in results', max_length=255, verbose_name='Name')),
                ('question', ckeditor.fields.RichTextField(help_text='as shown on evaluation form', verbose_name='Question')),
                ('item_type', models.CharField(choices=[('free_text', 'Free text'), ('agreement', 'Agreement')], max_length=255, verbose_name='Question type')),
                ('part', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationpart', verbose_name='Evaluation part')),
            ],
            options={
                'verbose_name': 'Evaluation question',
                'verbose_name_plural': 'Evaluation questions',
            },
        ),
        migrations.RemoveField(
            model_name='evaluationresult',
            name='question',
        ),
        migrations.AddField(
            model_name='evaluationgroup',
            name='group_name',
            field=models.CharField(default='unknown', max_length=255, verbose_name='Group name'),
            preserve_default=False,
        ),
        migrations.DeleteModel(
            name='EvaluationQuestion',
        ),
        migrations.AddField(
            model_name='evaluationresult',
            name='item',
            field=models.ForeignKey(default=0, on_delete=django.db.models.deletion.CASCADE, to='evalu.evaluationitem', verbose_name='Question'),
            preserve_default=False,
        ),
        migrations.AddConstraint(
            model_name='evaluationitem',
            constraint=models.UniqueConstraint(fields=('part', 'order'), name='part_site_order_unique'),
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-20 10:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0065_calendarevent_update_timezone_constraints'),
        ('evalu', '0019_drop_sites_add_managed_by'),
    ]

    operations = [
        migrations.AddField(
            model_name='evaluationphase',
            name='evaluation_group_types',
            field=models.ManyToManyField(blank=True, null=True, to='core.grouptype', verbose_name='Groups with these types can evaluate persons'),
        ),
    ]



from django.utils.translation import gettext as _

from dynamic_preferences.preferences import Section
from dynamic_preferences.types import IntegerPreference, StringPreference

from aleksis.core.registries import site_preferences_registry

general = Section("evalu", verbose_name=_("EvaLU"))


@site_preferences_registry.register
class Place(StringPreference):
    """Place for legal consents."""

    section = general
    name = "place"
    default = ""
    verbose_name = _("Place (for legal consents)")


@site_preferences_registry.register
class MinimumNumberOfPersonsComparisonGroups(IntegerPreference):
    """Minimum number of persons needed for comparison groups."""

    section = general
    name = "number_of_persons_comparison_groups"
    default = 3
    verbose_name = _("Minimum number of persons needed for comparison groups")


@site_preferences_registry.register
class MinimumNumberOfResultsComparisonGroups(IntegerPreference):
    """Minimum number of results needed for comparison groups."""

    section = general
    name = "number_of_results_comparison_groups"
    default = 5
    verbose_name = _("Minimum number of results needed for comparison groups")


@site_preferences_registry.register
class MinimumNumberOfPersonsForResults(IntegerPreference):
    """Minimum number of persons needed for results."""

    section = general
    name = "number_of_persons_results"
    default = 5
    verbose_name = _("Minimum number of persons needed for results")



import pytest

from aleksis.apps.evalu.models import EvaluationPart, QuestionType

pytestmark = pytest.mark.django_db


def test_str():
    o = EvaluationPart.objects.create(name="Foo", order=1)
    assert str(o) == "Foo"


def test_item_str():
    o = EvaluationPart.objects.create(name="Foo", order=1)
    item = o.items.create(name="Bar", order=1, question="Bar?", item_type=QuestionType.AGREEMENT)
    assert str(item) == "Bar"



# Generated by Django 4.2.6 on 2023-10-23 18:53

import aleksis.core.managers
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("evalu", "0013_remove_evaluationphase_results_date_start_site"),
    ]

    operations = [
        migrations.AddField(
            model_name="evaluationitem",
            name="polymorphic_ctype",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="polymorphic_%(app_label)s.%(class)s_set+",
                to="contenttypes.contenttype",
            ),
        ),
    ]



# Generated by Django 3.2.6 on 2021-09-15 16:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('evalu', '0002_evaluation_item'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='evaluationitem',
            options={'verbose_name': 'Evaluation item', 'verbose_name_plural': 'Evaluation items'},
        ),
        migrations.AlterField(
            model_name='evaluationitem',
            name='question',
            field=models.TextField(help_text='as shown on evaluation form', verbose_name='Question'),
        ),
    ]



# Generated by Django 3.2.9 on 2021-12-06 13:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0020_pdf_file_person_optional'),
        ('evalu', '0008_comparison_group'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='evaluationgroup',
            managers=[
            ],
        ),
        migrations.AlterField(
            model_name='comparisongroup',
            name='groups',
            field=models.ManyToManyField(related_name='evalu_comparison_groups', to='core.Group', verbose_name='Groups'),
        ),
        migrations.CreateModel(
            name='DoneEvaluationComparison',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('comparison_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='done_evaluations_comparison', to='evalu.comparisongroup', verbose_name='Comparison group')),
                ('evaluation_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='done_evaluations_comparison', to='evalu.evaluationgroup', verbose_name='Evaluation group')),
            ],
            options={
                'verbose_name': 'Done evaluation (comparison)',
                'verbose_name_plural': 'Done evaluations (comparison)',
            },
        ),
    ]
